/*  vtbanner.c
 *
 *  Simple program to print banner messages using VT100 type alternate
 *  character set.  Works well with the Linux console driver too.
 *
 *  I use this program to enhance boring login issue files. Very spiffy!
 *
 *   Copyright January 1995  Simon Allen, Coventry, UK.
 *  E-Mail : simonallen@cix.compulink.co.uk
 *
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2, or (at your option) any
 *  later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <ctype.h>
#include <stdio.h>
#include <string.h>

/* Change to alternate character set */
#define ALT "(0"

/* Change back to normal character set */
#define NORM "(B"

/* Inter character space */
#define SPCS " "

/* Width of character */
#define WIDTH 5

/* Height of character */
#define HEIGHT 5


/* The following define the characters. */
/* You may in all likelyhood be able to improve on these ;-) */
/* Please mail me the changes. */

/* Definition of numbers */
char *nums[] = {
	"lqqqkx   xx   xx   xmqqqj",   /* 0 */
	"  qk    x    x    x   qvq",   /* 1 */
	"lqqqk    xlqqqjx    mqqqj",   /* 2 */
	"lqqqk    x qqqu    xmqqqj",   /* 3 */
	"l   kx   xmqqqu    x    j",   /* 4 */
	"lqqqkx    mqqqk    xmqqqj",   /* 5 */
	"lqqqkx    tqqqkx   xmqqqj",   /* 6 */
	"lqqqk    x    x    x    j",   /* 7 */
	"lqqqkx   xtqqqux   xmqqqj",   /* 8 */
	"lqqqkx   xmqqqu    xmqqqj" }; /* 9 */

/* Definition of letters */
char *chars[] = {
	"lqqqkx   xtqqqux   xm   j",  /* A */
	"lqqk x  x tqqvkx   xmqqqj",  /* B */
	"lqqqkx    x    x    mqqqj",  /* C */
	"lqqk x  mkx   xx  ljmqqj ",  /* D */
	"lqqqkx    tqqq x    mqqqj",  /* E */
	"lqqqkx    tqqq x    m    ",  /* F */
	"lqqqkx    x  qkx   xmqqqj",  /* G */
	"l   kx   xtqqqux   xm   j",  /* H */
	" qwq   x    x    x   qvq ",  /* I */
	" qwq   x    x    x  mqj  ",  /* J */
	"l   kx   xtqqwjx  mkm   j",  /* K */
	"l    x    x    x    mqqqj",  /* L */
	"lqwqkx x xx x xx   xm   j",  /* M */
	"lqqqkx   xx   xx   xm   j",  /* N */
	"lqqqkx   xx   xx   xmqqqj",  /* O */
	"lqqqkx   xtqqqjx    m    ",  /* P */
	"lqqqkx   xx   xx   nmqqqj",  /* Q */
	"lqqqkx   xtqqwjx  mkm   j",  /* R */
	"lqqqkx    mqqqk    xmqqqj",  /* S */
	"lqwqk  x    x    x    v  ",  /* T */
	"l   kx   xx   xx   xmqqqj",  /* U */
	"l   kx   xx   xmk lj mqj ",  /* V */
	"l   kx   xx x xx x xmqvqj",  /* W */
	"l   kmk lj tqu lj mkm   j",  /* X */
	"l   kmk lj mwj   x    v  ",  /* Y */
	"lqqqk   lj lqj lj   mqqqj",};/* Z */

/* Definition of anything else you fancy... */
char *symbols[] = {
	"lqqqkx   x  lqj  x    f  "}; /* ? */


main(int argc, char *argv[])
{
	char message[1024];
	int j;

	if(argc<2)
	{
		printf("Enter string : ");
		gets(message);
	
		print_string(message);
	}
	else
	if(!strncmp(argv[1], "-h", 2))
	{
		fprintf(stderr, "Usage : %s [message to print]\n", argv[0]);
		exit(1);
	}
	else
	{
		strcpy(message, "");
		/* Join command line args together */
		for(j=1; j<argc; j++)
		{
			strcat(message, argv[j]);
			strcat(message, " ");
		}

		print_string(message);
	}
}

print_string(char *s)
{
	int line, j, c, k;

	for(line=0; line<HEIGHT; line++)
	{
		printf("%s", ALT);
		for(j=0; j<strlen(s); j++)
		{
			c = toupper(s[j]);
			if(isdigit(c))
			{
				for(k=line*WIDTH; k<line*WIDTH+WIDTH; k++)
					printf("%c", *(nums[(c-'0')]+k));
				printf("%s", SPCS);
			}
			else
			if(isupper(c))
			{
				for(k=line*WIDTH; k<line*WIDTH+WIDTH; k++)
					printf("%c", *(chars[(c-'A')]+k));
				printf("%s", SPCS);
			}
			else /* Space */
			if(c==32)
			{
				printf("%s%s%s", SPCS, SPCS, SPCS);
			}
			else /* Anything else */
			{
				for(k=line*WIDTH; k<line*WIDTH+WIDTH; k++)
					printf("%c", *(symbols[0]+k));
				printf("%s", SPCS);
			}
		}
		printf("%s", NORM);
		printf("\n");
	}
}
